@@ -24,6 +24,7 @@ from urllib.parse import parse_qs, urlparse
 from asyncopenstackclient import AuthPassword, GlanceClient
 from simple_rest_client.exceptions import NotFoundError, ServerError
 
+from mrack.context import global_context
 from mrack.errors import (
     NotAuthenticatedError,
     ProviderError,
@@ -61,6 +62,7 @@ class OpenStackProvider(Provider):
 
     def __init__(self):
         """Object initialization."""
+        super().__init__()
         self._name = PROVISIONER_KEY
         self.dsp_name = "OpenStack"
         self.max_retry = 1  # provisioning retries
@@ -88,7 +90,7 @@ class OpenStackProvider(Provider):
             # https://docs.openstack.org/api-guide/compute/server_concepts.html
         }
 
-    async def _opentack_gather_responses(self, *functions):
+    async def _openstack_gather_responses(self, *functions):
         """Gather the async result of functions from parameters.
 
         Returns:
@@ -155,12 +157,12 @@ class OpenStackProvider(Provider):
         self.network_pools = networks
         object_start = datetime.now()
 
-        _, _, self.limits, _, _ = await self._opentack_gather_responses(
-            self.load_flavors(),
-            self.load_images(image_names),
+        _, _, self.limits, _, _ = await self._openstack_gather_responses(
+            self._load_flavors(),
+            self._load_images(image_names),
             self.nova.limits.show(),
-            self.load_networks(),
-            self.load_ip_availabilities(),
+            self._load_networks(),
+            self._load_ip_availabilities(),
         )
 
         object_duration = datetime.now() - object_start
@@ -240,11 +242,11 @@ class OpenStackProvider(Provider):
         Full capacity of these 5 nets: 500
 
         Normalized network range:
-        net 4 - <0, 0.29>
-        net 3 - (0,29, 0.55>
-        net 5 - (0.55, 0.76>
-        net 2 - (0.76, 0.96>
-        net 1 - (0.96, 1>
+        net 4 - <0, 0.29)
+        net 3 - <0,29, 0.55)
+        net 5 - <0.55, 0.76)
+        net 2 - <0.76, 0.96)
+        net 1 - <0.96, 1)
 
         Which will divide networks for hosts:
         host 1 => net 4 (0 falls into net 4 interval)
@@ -257,7 +259,7 @@ class OpenStackProvider(Provider):
         """
         chosen = tuple()
         for index, net in enumerate(nets_weights):
-            if host_weight <= net[NETWORK_SIZE]:
+            if host_weight < net[NETWORK_SIZE]:
                 chosen = net
                 # print details only if there is more one network to be picked from
                 if len(nets_weights) > 1:
@@ -285,47 +287,107 @@ class OpenStackProvider(Provider):
 
         Available networks with most IPs available is picked.
         """
-        possible_nets = self.network_pools[network_type]
-        # filter out the None values - the get_network return None if network not found
-        networks = list(filter(None, [self.get_network(net) for net in possible_nets]))
-        usable = []
-        low_avail_nets = []
+        usable = set()
+        low_avail_nets = set()
         total_available = 0
-        for network in networks:
-            if not network:
-                continue
+        max_size_net = 0
+        threshold = global_context.CONFIG.usable_network_threshold
+        spread_option = global_context.CONFIG.network_spread
+        big_request = False
 
-            ips = self.get_ips(ref=network.get("id"))
+        # filter out the None values - the get_network return None if network not found
+        for network in list(
+            filter(
+                None,
+                [self._get_network(net) for net in self.network_pools[network_type]],
+            )
+        ):
+            ips = self._get_ips(ref=network.get("id"))
 
             available = ips["total_ips"] - ips["used_ips"]
-            logger.debug(f"{self.dsp_name} Network: {network['name']}")
+            max_size_net = available if available > max_size_net else max_size_net
+
+            logger.debug(
+                f"{self.dsp_name} Network: {network['name']}"
+                f"{' - unusable (0 IPs left) skipping' if not available else ''}"
+            )
             logger.debug(f"  total: {ips['total_ips']}")
             logger.debug(f"  used: {ips['used_ips']}")
             logger.debug(f"  available: {available}")
 
-            total_available += available
-
-            #  assume network is unusable when < 5% IPs left - (12/251) when IPv4
-            if available / ips["total_ips"] < 0.05:
-                low_avail_nets.append((network["name"], available))
+            if not available:
                 continue
 
-            usable.append((network["name"], available))
+            total_available += available
+
+            #  assume network is unusable when usable_threshold is reached
+            if ips["used_ips"] / ips["total_ips"] > threshold / 100:
+                low_avail_nets.add((network["name"], available))
+                continue
+
+            usable.add((network["name"], available))
 
         if not usable and (total_available >= requested_ip_cnt) and low_avail_nets:
-            usable += low_avail_nets
+            usable = usable.union(low_avail_nets)
 
         if not usable:
             raise ValidationError(
-                "Error: no available networks for "
+                "No available networks for "
                 f"{requested_ip_cnt} hosts with {network_type}",
                 self.dsp_name,
             )
 
-        if len(usable) > requested_ip_cnt:
-            usable = sample(usable, requested_ip_cnt)
+        if requested_ip_cnt > max_size_net:
+            big_request = True
+            # if we request more IPs than maximum size net consider the request big
+            # adding even small networks to the usable set to spread the load
+            usable = usable.union(low_avail_nets)
 
-        return sorted(usable, key=lambda u: u[NETWORK_SIZE], reverse=True)
+        logger.debug(
+            f"{self.dsp_name} Default mrack behavior is to 'allow' spread"
+            " the load to all available networks, to disable this feature set"
+            " the 'network-spread' option in the mrack.conf file to 'no'"
+        )
+        logger.debug(
+            f"{self.dsp_name} 'network-spread' option is set to '{spread_option}'"
+        )
+        logger.debug(
+            f"{self.dsp_name} Considering this request "
+            f"{'big' if big_request else 'small'} - requesting {requested_ip_cnt} "
+            f"IP(s) from total available {total_available} IP(s)"
+        )
+
+        size_for_spread = (
+            len(usable) if len(usable) <= requested_ip_cnt else requested_ip_cnt
+        )
+
+        # number of networks which we will consider to pass to picking algorithm
+        network_subset_size = 0
+
+        if spread_option == "force":
+            network_subset_size = size_for_spread
+        elif spread_option == "no":
+            if not big_request:
+                network_subset_size = 1
+        else:  # spread option is allow or some unknown string -> fallback to default
+            if usable.intersection(low_avail_nets) or big_request:
+                network_subset_size = size_for_spread
+            else:
+                network_subset_size = 1
+
+        if network_subset_size:
+            return sorted(
+                set(sample(sorted(usable), network_subset_size)),
+                key=lambda u: u[NETWORK_SIZE],
+                reverse=True,
+            )
+        # If the above return does not happened because there is 0 usable networks
+        # we will raise a ValidationError
+        raise ValidationError(
+            f"Can not satisfy request for {requested_ip_cnt} hosts ({network_type}) "
+            "Change the 'network-spread' in mrack.conf or try provisioning again later",
+            self.dsp_name,
+        )
 
     def _get_weights_from_usable(self, usable_networks):
         """
@@ -359,7 +421,7 @@ class OpenStackProvider(Provider):
 
         return weights
 
-    def translate_network_types(self, hosts):
+    def _translate_network_types(self, hosts):
         """Pick the right OpenStack networks for all hosts.
 
         Pick the network based on network type, networks configured for the
@@ -372,6 +434,7 @@ class OpenStackProvider(Provider):
         nt_requirements = self._aggregate_networks(hosts)
         nt_map = {}
         for network_type, count in nt_requirements.items():
+            # count is crucial
             nt_map[network_type] = self._get_usable_networks(network_type, count)
 
         weight_map = {}
@@ -401,6 +464,7 @@ class OpenStackProvider(Provider):
                 continue
 
             networks_weights = weight_map[network_type]
+
             host["network"] = self._pick_network(
                 host_name=host["name"],
                 host_weight=hosts.index(host) / len(hosts),
@@ -413,21 +477,21 @@ class OpenStackProvider(Provider):
             self.networks[network["name"]] = network
             self.networks_by_ref[network["id"]] = network
 
-    def get_flavor(self, name=None, ref=None):
+    def _get_flavor(self, name=None, ref=None):
         """Get flavor by name or UUID."""
         flavor = self.flavors.get(name)
         if not flavor:
             flavor = self.flavors_by_ref.get(ref)
         return flavor
 
-    def get_image(self, name=None, ref=None):
+    def _get_image(self, name=None, ref=None):
         """Get image by name or UUID."""
         image = self.images.get(name)
         if not image:
             image = self.images_by_ref.get(ref)
         return image
 
-    def get_network(self, name=None, ref=None):
+    def _get_network(self, name=None, ref=None):
         """Get network by name or UUID."""
         network = self.networks.get(name)
         if not network:
@@ -441,21 +505,21 @@ class OpenStackProvider(Provider):
 
         return network
 
-    def get_ips(self, name=None, ref=None):
+    def _get_ips(self, name=None, ref=None):
         """Get network availability by network name or network UUID."""
         aval = self.ips.get(name)
         if not aval:
             aval = self.ips_by_ref.get(ref)
         return aval
 
-    async def load_flavors(self):
+    async def _load_flavors(self):
         """Extend provider configuration by loading all flavors from OpenStack."""
         resp = await self.nova.flavors.list()
         flavors = resp["flavors"]
         self._set_flavors(flavors)
         return flavors
 
-    async def load_images(self, image_names=None):
+    async def _load_images(self, image_names=None):
         """
         Extend provider configuration by loading information about images.
 
@@ -489,14 +553,14 @@ class OpenStackProvider(Provider):
 
         return images
 
-    async def load_networks(self):
+    async def _load_networks(self):
         """Extend provider configuration by loading all networks from OpenStack."""
         resp = await self.neutron.network.list()
         networks = resp["networks"]
         self._set_networks(networks)
         return networks
 
-    async def load_ip_availabilities(self):
+    async def _load_ip_availabilities(self):
         """Extend provider configuration by loading networks availabilities."""
         resp = await self.neutron.ip.list()
         availabilities = resp["network_ip_availabilities"]
@@ -510,9 +574,9 @@ class OpenStackProvider(Provider):
         flavor_ref = req.get("flavorRef")
         flavor = None
         if flavor_ref:
-            flavor = self.get_flavor(ref=flavor_ref)
+            flavor = self._get_flavor(ref=flavor_ref)
         if flavor_spec:
-            flavor = self.get_flavor(flavor_spec, flavor_spec)
+            flavor = self._get_flavor(flavor_spec, flavor_spec)
 
         if not flavor:
             specs = f"flavor: {flavor_spec}, ref: {flavor_ref}"
@@ -524,9 +588,9 @@ class OpenStackProvider(Provider):
         image_ref = req.get("imageRef")
         image = None
         if image_ref:
-            image = self.get_image(ref=image_ref)
+            image = self._get_image(ref=image_ref)
         if image_spec:
-            image = self.get_image(image_spec, image_spec)
+            image = self._get_image(image_spec, image_spec)
         if not image:
             specs = f"image: {image_spec}, ref: {image_ref}"
             raise ValidationError(f"Image not found {specs}")
@@ -541,12 +605,12 @@ class OpenStackProvider(Provider):
             network_specs = []
         for network_spec in network_specs:
             uuid = network_spec.get("uuid")
-            network = self.get_network(ref=uuid)
+            network = self._get_network(ref=uuid)
             if not network:
                 raise ValidationError(f"Network not found: {network_spec}")
             networks.append(network)
         if network_req:
-            network = self.get_network(name=network_req, ref=network_req)
+            network = self._get_network(name=network_req, ref=network_req)
             if not network:
                 raise ValidationError(f"Network not found: {network_req}")
 
@@ -595,7 +659,7 @@ class OpenStackProvider(Provider):
         if prepare_images:
             im_list = ", ".join(prepare_images)
             logger.debug(f"{self.dsp_name} Loading image info for: '{im_list}'")
-            await self.load_images(list(prepare_images))
+            await self._load_images(list(prepare_images))
             logger.debug(f"{self.dsp_name} Loading images info done.")
 
         self._set_poll_sleep_times(reqs)
@@ -604,7 +668,7 @@ class OpenStackProvider(Provider):
     async def validate_hosts(self, reqs):
         """Validate that all hosts requirements contains existing required objects."""
         # translate network type to actual network and check network availabilities
-        self.translate_network_types(reqs)
+        self._translate_network_types(reqs)
 
         for req in reqs:
             logger.info(f"{self.dsp_name} Validating host: {object2json(req)}")
@@ -617,9 +681,9 @@ class OpenStackProvider(Provider):
         flavor_ref = req.get("flavorRef")
         flavor = None
         if flavor_ref:
-            flavor = self.get_flavor(ref=flavor_ref)
+            flavor = self._get_flavor(ref=flavor_ref)
         if flavor_spec:
-            flavor = self.get_flavor(flavor_spec, flavor_spec)
+            flavor = self._get_flavor(flavor_spec, flavor_spec)
 
         try:
             res = {"ram": flavor["ram"], "vcpus": flavor["vcpus"]}
@@ -648,7 +712,7 @@ class OpenStackProvider(Provider):
 
         # poll the actual openstack load
         logger.debug(f"{self.dsp_name} Loading nova limits")
-        limits_await = await self._opentack_gather_responses(self.nova.limits.show())
+        limits_await = await self._openstack_gather_responses(self.nova.limits.show())
         self.limits = limits_await[0]  # gather returns list
 
         limits = self.limits["limits"]["absolute"]
@@ -672,6 +736,18 @@ class OpenStackProvider(Provider):
 
         return req_vcpus <= limit_vcpus and req_memory <= limit_memory
 
+    async def utilization(self):
+        """Check utilization of provider."""
+        limits_await = await self._openstack_gather_responses(self.nova.limits.show())
+        self.limits = limits_await[0]  # gather returns list
+        used_vcpus = self.limits["totalCoresUsed"]
+        used_memory = self.limits["totalRAMUsed"]
+        limit_vcpus = self.limits["maxTotalCores"]
+        limit_memory = self.limits["maxTotalRAMSize"]
+        cpu_util = used_vcpus / limit_vcpus * 100
+        memory_util = used_memory / limit_memory * 100
+        return cpu_util if memory_util <= cpu_util else memory_util
+
     async def create_server(self, req):
         """Issue creation of a server.
 
